from .algorithm import Algorithm as Algorithm
from .dataset import Dataset as Dataset
from .evaluation import Evaluation as Evaluation
from .injection import Injection as Injection
from .task import Task as Task



# Run this file:
# uv run pytest -s tests/test_algorithm_api.py
from rcabench.model.common import SubmitResult
from pprint import pprint
import pytest


@pytest.mark.parametrize(
    "payloads",
    [
        (
            [
                {
                    "image": "e-diagnose",
                    "dataset": "ts-ts-ui-dashboard-pod-failure-c27jzh",
                }
            ]
        ),
        (
            [
                {
                    "image": "rcabench-rcaeval-generic",
                    "dataset": "ts-ts-travel-service-delay-ftz7lb",
                    "env_vars": {"ALGORITHM": "baro", "VENV": "default"},
                },
                {
                    "image": "rcabench-rcaeval-generic",
                    "dataset": "ts-ts-travel-service-delay-ftz7lb",
                    "env_vars": {"ALGORITHM": "nsigma", "VENV": "default"},
                },
                {
                    "image": "rcabench-rcaeval-generic",
                    "dataset": "ts-ts-travel-service-delay-ftz7lb",
                    "env_vars": {"ALGORITHM": "circa", "VENV": "default"},
                },
                {
                    "image": "rcabench-rcaeval-generic",
                    "dataset": "ts-ts-travel-service-delay-ftz7lb",
                    "env_vars": {"ALGORITHM": "rcd", "VENV": "rcd"},
                },
            ]
        ),
    ],
)
def test_submit_algorithms(sdk, payloads):
    """测试批量提交算法"""
    resp = sdk.algorithm.submit(payloads)
    pprint(resp)

    if not isinstance(resp, SubmitResult):
        pytest.fail(resp.model_dump_json())

    traces = resp.traces
    if not traces:
        pytest.fail("No traces returned from execution")



from typing import AsyncGenerator, List, Optional
from ..client.async_client import AsyncSSEClient, ClientManager
from ..model.task import QueueItem
from contextlib import asynccontextmanager
from uuid import UUID
import aiohttp
import asyncio

__all__ = ["Stream"]


class Stream:
    CLIENT_NAME = "SSE-{client_id}"

    def __init__(self, base_url: str, max_connections: int = 10):
        self.base_url = base_url
        self.client_manager = ClientManager()
        self.conn_pool = asyncio.Queue(max_connections)
        self.active_connections = set()
        self.index = -1

    @asynccontextmanager
    async def _get_session(self) -> AsyncGenerator[aiohttp.ClientSession, None]:
        session = await self.conn_pool.get()
        try:
            yield session
        finally:
            await self.conn_pool.put(session)

    async def _stream_client(
        self,
        index: int,
        client_id: UUID,
        url: str,
        client_timeout: float,
        to_client_manager: bool,
    ) -> Optional[QueueItem]:
        retries = 0
        max_retries = 3

        sse_client = AsyncSSEClient(
            client_id,
            f"{self.base_url}{url}",
            self.client_manager if to_client_manager else None,
        )
        self.active_connections.add(client_id)

        while retries < max_retries:
            try:
                return await sse_client.connect(client_timeout * (index + 1))
            except aiohttp.ClientError:
                retries += 1
                await asyncio.sleep(2**retries)
            finally:
                await sse_client.close()
                self.active_connections.discard(client_id)

    async def add_result_queue(self) -> asyncio.Queue[QueueItem]:
        queue = asyncio.Queue()
        self.client_manager.result_queue = queue
        return queue

    async def start_single_stream(
        self, client_id: UUID, url: str, client_timeout: float
    ) -> QueueItem:
        self.index += 1
        return await self._stream_client(
            self.index,
            client_id,
            url,
            client_timeout,
            to_client_manager=False,
        )

    async def start_multiple_stream(
        self, client_ids: List[UUID], urls: List[str], client_timeout: float
    ) -> None:
        """批量启动多个SSE流"""
        for idx, (client_id, url) in enumerate(zip(client_ids, urls)):
            self.index += idx + 1
            asyncio.create_task(
                self._stream_client(
                    self.index,
                    client_id,
                    url,
                    client_timeout,
                    to_client_manager=True,
                ),
                name=self.CLIENT_NAME.format(client_id=client_id),
            )

    async def stop_stream(self, client_id: UUID):
        """停止指定SSE流"""
        for task in asyncio.all_tasks():
            if task.get_name() == self.CLIENT_NAME.format(client_id=client_id):
                task.cancel()
                break

    async def stop_all_streams(self):
        """停止所有SSE流"""
        for client_id in list(self.active_connections):
            await self.stop_stream(client_id)

    async def cleanup(self):
        """清理所有资源"""
        await self.stop_all_streams()
        while not self.conn_pool.empty():
            session = await self.conn_pool.get()
            await session.close()
        await self.client_manager.cleanup()



from typing import List, Optional
from ..const import TIME_EXAMPLE, TIME_FORMAT
from datetime import datetime, timedelta, timezone
from pydantic import BaseModel, Field


class ModelHTTPError(BaseModel):
    """
    HTTP标准错误响应模型

    Attributes:
        status_code: HTTP状态码
        detail: 错误详情信息
        timestamp: 错误发生时间（带时区偏移）
        path: 请求路径（可选）
        method: HTTP请求方法（可选）
    """

    status_code: int = Field(
        ...,
        description="HTTP status code indicating the error type",
        json_schema_extra={"example": 404},
    )

    detail: str = Field(
        ...,
        description="Human-readable error message explaining what went wrong",
        json_schema_extra={"example": "Item not found"},
    )

    timestamp: str = Field(
        default_factory=lambda: datetime.now(timezone(timedelta(hours=8))).strftime(
            TIME_FORMAT
        ),
        description="ISO 8601 format time (with time zone offset)",
        json_schema_extra={"example": TIME_EXAMPLE},
    )

    path: Optional[str] = Field(
        None,
        description="API endpoint path that triggered the error",
        json_schema_extra={"example": "/api/items/123"},
    )

    method: Optional[str] = Field(
        None,
        description="HTTP method used for the request",
        json_schema_extra={"example": "GET"},
    )


class FieldValidationIssue(BaseModel):
    """
    单个字段验证错误详情

    Attributes:
        field: 发生错误的字段路径
        message: 具体错误描述
        error_type: 错误类型标识
    """

    field: str = Field(
        ...,
        description="Dot-path to the invalid field",
        json_schema_extra={"example": "user_id"},
    )

    message: str = Field(
        ...,
        description="Human-readable error message",
        json_schema_extra={"example": "Value is not a valid UUID"},
    )

    error_type: str = Field(
        ...,
        description="Error type identifier",
        json_schema_extra={"example": "value_error.uuid"},
    )


class ModelValidationError(BaseModel):
    """
    模型验证错误响应模型

    Attributes:
        detail: 错误信息
        issuse: 详细错误列表
        timestamp 错误发生时间（带时区偏移）
    """

    detail: str = Field(
        ...,
        description="Error Summary",
    )

    issues: List[FieldValidationIssue] = Field(
        ...,
        description="Detailed list of validation errors",
        min_length=1,
    )

    timestamp: str = Field(
        default_factory=lambda: datetime.now(timezone(timedelta(hours=8))).strftime(
            TIME_FORMAT
        ),
        description="ISO 8601 format time (with time zone offset)",
        json_schema_extra={"example": TIME_EXAMPLE},
    )



# Run this file:
# uv run pytest -s tests/test_task_api.py
from typing import Any, Dict, List, Optional
from conftest import BASE_URL
from pprint import pprint
from rcabench.logger import logger
from rcabench.model.common import SubmitResult
from rcabench.model.task import QueueItem
from rcabench.rcabench import RCABenchSDK
from uuid import UUID
import asyncio
import pytest


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "benchmark, interval, pre_duration, specs, max_items_per_consumer",
    [
        # One spec
        (
            "clickhouse",
            3,
            1,
            [
                {
                    "children": {
                        "1": {
                            "children": {
                                "0": {"value": 1},
                                "1": {"value": 0},
                                "2": {"value": 42},
                            }
                        },
                    },
                    "value": 1,
                },
            ],
            1,
        ),
        # Many specs
        (
            "clickhouse",
            3,
            1,
            [
                {
                    "children": {
                        "1": {
                            "children": {
                                "0": {"value": 1},
                                "1": {"value": 0},
                                "2": {"value": 42},
                            }
                        },
                    },
                    "value": 1,
                },
                {
                    "children": {
                        "4": {
                            "children": {
                                "0": {"value": 1},
                                "1": {"value": 0},
                                "2": {"value": 26},
                                "3": {"value": 10},
                                "4": {"value": 2},
                            }
                        },
                    },
                    "value": 4,
                },
                {
                    "children": {
                        "1": {
                            "children": {
                                "0": {"value": 1},
                                "1": {"value": 0},
                                "2": {"value": 42},
                            }
                        },
                    },
                    "value": 1,
                },
            ],
            1,
        ),
        # Total timeout
        (
            "clickhouse",
            3,
            1,
            [
                {
                    "children": {
                        "1": {
                            "children": {
                                "0": {"value": 1},
                                "1": {"value": 0},
                                "2": {"value": 42},
                            }
                        },
                    },
                    "value": 1,
                },
            ],
            1,
        ),
    ],
)
async def test_injection_and_building_dataset_batch(
    benchmark: str,
    interval: int,
    pre_duration: int,
    specs: List[Dict[str, Any]],
    max_items_per_consumer: int,
):
    per_consumer_timeout = (interval + 1) * 60
    total_timeout = len(specs) * (interval + 1) * 60

    results = {}
    try:
        results = await asyncio.wait_for(
            injection_and_building_dataset_batch(
                benchmark,
                interval,
                pre_duration,
                specs,
                max_items_per_consumer,
                per_consumer_timeout,
            ),
            timeout=total_timeout,
        )
    except asyncio.TimeoutError:
        logger.error("Total time out")
    finally:
        logger.info(f"Final results: {results}")


async def injection_and_building_dataset_batch(
    benchmark: str,
    interval: int,
    pre_duration: int,
    specs: List[Dict[str, Any]],
    max_items_per_consumer: int,
    per_consumer_timeout: float,
):
    sdk = RCABenchSDK(BASE_URL)

    resp = sdk.injection.submit(benchmark, interval, pre_duration, specs)
    logger.info(resp)

    traces = resp.traces

    task_ids = [trace.head_task_id for trace in traces]
    trace_ids = [trace.trace_id for trace in traces]
    queue = await sdk.task.get_stream_batch(task_ids, trace_ids, interval * 60)

    num_consumers = len(traces) // max_items_per_consumer
    results = await run_consumers(
        queue,
        num_consumers,
        max_items_per_consumer,
        per_consumer_timeout,
    )

    await sdk.task.stream.cleanup()
    return results


async def run_consumers(
    queue: asyncio.Queue[QueueItem],
    num_consumers: int,
    max_items_per_consumer: int,
    per_consumer_timeout: float,
):
    all_results = {}

    for i in range(num_consumers):
        all_results[f"Batch-{i}"] = await consumer_task(
            i,
            queue,
            max_items_per_consumer,
            per_consumer_timeout,
        )

    return all_results


async def consumer_task(
    consumer_id: int,
    queue: asyncio.Queue[QueueItem],
    max_num: int,
    timeout: float,
) -> Optional[List[Dict[str, Any]]]:
    try:
        results = await consumer(queue, max_num, timeout)
        logger.info(f"Consumer-{consumer_id} completed")
        return results
    except asyncio.TimeoutError:
        logger.error(f"Consumer-{consumer_id} timed out")
        return None
    except Exception as e:
        logger.error(f"Consumer-{consumer_id} failed: {str(e)}")
        return None


async def consumer(
    queue: asyncio.Queue[QueueItem],
    max_num: int,
    timeout: float,
) -> List[Dict[str, Any]]:
    results = []
    count = 0
    while count < max_num:
        try:
            item = await asyncio.wait_for(queue.get(), timeout)
            results.append(item.model_dump(exclude_unset=True))
            count += 1
        except asyncio.TimeoutError:
            logger.warning("Timeout while waiting for queue item")
            raise

    return results


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "benchmark, interval, pre_duration, specs",
    [
        (
            "clickhouse",
            2,
            1,
            [
                {
                    "children": {
                        "1": {
                            "children": {
                                "0": {"value": 1},
                                "1": {"value": 0},
                                "2": {"value": 42},
                            }
                        },
                    },
                    "value": 1,
                }
            ],
        ),
        (
            "clickhouse",
            2,
            0,
            [
                {
                    "children": {
                        "1": {
                            "children": {
                                "0": {"value": 1},
                                "1": {"value": 0},
                                "2": {"value": 42},
                            }
                        },
                    },
                    "value": 1,
                }
            ],
        ),
        (
            "clickhouse",
            2,
            1,
            [
                {
                    "children": {
                        "1": {
                            "children": {
                                "0": {"value": -1},
                                "1": {"value": 0},
                                "2": {"value": 42},
                            }
                        },
                    },
                    "value": 1,
                }
            ],
        ),
    ],
)
async def test_injection_and_building_dataset_all(
    benchmark, interval, pre_duration, specs
):
    sdk = RCABenchSDK(BASE_URL)

    resp = sdk.injection.submit(benchmark, interval, pre_duration, specs)
    pprint(resp)

    if not isinstance(resp, SubmitResult):
        pytest.fail(resp.model_dump_json())

    traces = resp.traces
    if not traces:
        pytest.fail("No traces returned from execution")

    task_ids = [trace.head_task_id for trace in traces]
    trace_ids = [trace.trace_id for trace in traces]
    report = await sdk.task.get_stream_all(task_ids, trace_ids, timeout=None)
    report = report.model_dump(exclude_unset=True)
    pprint(report)

    return report


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "benchmark, interval, pre_duration, specs",
    [
        (
            "clickhouse",
            3,
            1,
            [
                {
                    "children": {
                        "16": {
                            "children": {
                                "0": {"value": 1},
                                "1": {"value": 0},
                                "2": {"value": 10},
                                "3": {"value": -576},
                            }
                        },
                    },
                    "value": 16,
                },
            ],
        ),
    ],
)
async def test_injection_and_building_dataset_single(
    benchmark, interval, pre_duration, specs
):
    sdk = RCABenchSDK(BASE_URL)

    if len(specs) != 1:
        pytest.fail("The length of specs must be 1")

    resp = sdk.injection.submit(benchmark, interval, pre_duration, specs)
    pprint(resp)

    if not isinstance(resp, SubmitResult):
        pytest.fail(resp.model_dump_json())

    traces = resp.traces
    if not traces:
        pytest.fail("No traces returned from execution")

    task_id = [trace.head_task_id for trace in traces][0]
    trace_id = [trace.trace_id for trace in traces][0]

    timeout = (interval + 1) * 60
    report = await sdk.task.get_stream_single(task_id, trace_id, timeout)
    report = report.model_dump(exclude_unset=True)
    pprint(report)

    await sdk.task.stream.cleanup()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "payload",
    [
        (
            [
                {
                    "algorithm": "e-diagnose",
                    "dataset": "ts-ts-travel2-service-pod-failure-rkxslq",
                }
            ]
        )
    ],
)
async def test_execute_algorithm_and_collection(payload: List[Dict[str, str]]):
    """测试执行多个算法并验证结果流收集功能

    验证步骤：
    1. 初始化 SDK 连接
    2. 获取可用算法列表
    3. 为每个算法生成执行参数
    4. 提交批量执行请求
    5. 启动流式结果收集
    6. 验证关键结果字段
    """
    sdk = RCABenchSDK(BASE_URL)

    resp = sdk.algorithm.submit(payload)
    pprint(resp)

    if not isinstance(resp, SubmitResult):
        pytest.fail(resp.model_dump_json())

    traces = resp.traces
    if not traces:
        pytest.fail("No traces returned from execution")

    task_ids = [trace.head_task_id for trace in traces]
    trace_ids = [trace.trace_id for trace in traces]
    report = await sdk.task.get_stream_all(task_ids, trace_ids, timeout=30)
    report = report.model_dump(exclude_unset=True)
    pprint(report)

    return report


@pytest.mark.asyncio
async def test_workflow():
    injection_payload = {
        "benchmark": "clickhouse",
        "interval": 2,
        "pre_duration": 1,
        "specs": [
            {
                "children": {
                    "1": {
                        "children": {
                            "0": {"value": 1},
                            "1": {"value": 0},
                            "2": {"value": 42},
                        }
                    },
                },
                "value": 1,
            }
        ],
    }

    injection_report = await test_injection_and_building_dataset_all(
        **injection_payload
    )
    datasets = extract_values(injection_report, "dataset")
    pprint(datasets)

    payload = []
    algorithms = ["e-diagnose"]
    for algorithm in algorithms:
        for dataset in datasets:
            payload.append({"algorithm": algorithm, "dataset": dataset})

    execution_report = await test_execute_algorithm_and_collection(payload)
    execution_ids = extract_values(execution_report, "execution_id")
    pprint(execution_ids)


def extract_values(data: Dict[UUID, Any], key: str) -> List[str]:
    """递归提取嵌套结构中的所有value值

    Args:
        data: 输入的嵌套字典结构，键可能为UUID

    Returns:
        所有找到的value值列表
    """
    values = []

    def _recursive_search(node):
        if isinstance(node, dict):
            # 检查当前层级是否有 key 字段
            if key in node:
                values.append(node[key])
            # 递归处理所有子节点
            for value in node.values():
                _recursive_search(value)
        elif isinstance(node, (list, tuple)):
            # 处理可迭代对象
            for item in node:
                _recursive_search(item)

    _recursive_search(data)
    return values
